/* Reflow Oven Controller
 *
 * Copyright (C) 2020  Mario Hüttel <mario.huettel@gmx.net>
 *
 * This file is part of the Reflow Oven Controller Project.
 *
 * The reflow oven controller is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * GDSII-Converter is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the reflow oven controller project.
 * If not, see <http://www.gnu.org/licenses/>.
 */

/**
 * @file systick.h
 */

#ifndef __SYSTICK_H__
#define __SYSTICK_H__

#include <stdint.h>
#include <stdbool.h>

/**
 * @brief Reload value for the systick timer.
 *
 * This value has to be configured to set the systick to a 100 us tick interval
 * The default value is 16800, which results in a 100us tick for 168 MHz CPU speed
 */
#define SYSTICK_RELOAD (16800UL)

/**
 * @brief Variable used by the systick_wait_ms function
 */
extern volatile uint32_t wait_tick_ms;

/**
 * @brief Systemclock in milliseconds.
 *
 * This value must not be reset during the whole runtime.
 * @warning In order to use this, you must assure that the read access is atomic.
 */
extern volatile uint64_t global_tick_ms;

/**
 * @brief Wait counter for the display. This must not be used anywhere else
 */
extern volatile uint32_t lcd_tick_100us;

/**
 * @brief Setup the Systick timer to generate a 100 us tick
 */
void systick_setup(void);

/**
 * @brief Wait for x milliseconds
 *
 * This function is not reentrant and must not be called from an interrupt
 *
 * @warning Do not use in interrupt context
 * @param ms wait time in ms
 */
void systick_wait_ms(uint32_t ms);

/**
 * @brief Get the gloabl millisecond tick. This can be used to implement all sorts of time based procedures / waits.
 * @warning Use this with care in interrupts. It may lead to race conditions. It is generally safe for use in standard program flow though.
 * @return  Global millisecond tick.
 */
uint64_t systick_get_global_tick(void);

/**
 * @brief Calculate the uptime from the current millisecond tick.
 * @param[out] days Days. May be NULL.
 * @param[out] hours Hours. May be NULL.
 * @param[out] minutes Minutes. May be NULL.
 * @param[out] seconds Seconds. May be NULL.
 */
void systick_get_uptime_from_tick(uint32_t *days, uint32_t *hours, uint32_t *minutes, uint32_t *seconds);

/**
 * @brief Check if \p ticks (milliseconds) have passed since \p start_timestamp
 * @param start_timestamp Start timestamp
 * @param ticks tick count
 * @return true: Time has passed
 */
bool systick_ticks_have_passed(uint64_t start_timestamp, uint64_t ticks);

#endif /* __SYSTICK_H__ */
